/*	$NetBSD: clnt_generic.c,v 1.18 2000/07/06 03:10:34 christos Exp $	*/

/*
 * The contents of this file are subject to the Sun Standards
 * License Version 1.0 the (the "License";) You may not use
 * this file except in compliance with the License.  You may
 * obtain a copy of the License at lib/libc/rpc/LICENSE
 *
 * Software distributed under the License is distributed on
 * an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 *
 * The Original Code is Copyright 1998 by Sun Microsystems, Inc
 *
 * The Initial Developer of the Original Code is:  Sun
 * Microsystems, Inc.
 *
 * All Rights Reserved.
 *
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 *
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/* #ident	"@(#)clnt_generic.c	1.40	99/04/21 SMI" */

#if defined(LIBC_SCCS) && !defined(lint)
static char *sccsid2 = "from: @(#)clnt_generic.c 1.4 87/08/11 (C) 1987 SMI";
static char *sccsid = "from: @(#)clnt_generic.c	2.2 88/08/01 4.0 RPCSRC";
#endif
#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * Copyright (c) 1986-1996,1998 by Sun Microsystems, Inc.
 * All rights reserved.
 */
#include "namespace.h"
#include "reentrant.h"
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <syslog.h>
#include <rpc/rpc.h>
#include <rpc/nettype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "un-namespace.h"
#include "rpc_com.h"
#ifdef NETAPP
#include "mt_misc.h"
#endif /* NETAPP */

extern bool_t __rpc_is_local_host(const char *);
int __rpc_raise_fd(int);

#ifndef NETIDLEN
#define	NETIDLEN 32
#endif


/*
 * Generic client creation with version checking the value of
 * vers_out is set to the highest server supported value
 * vers_low <= vers_out <= vers_high  AND an error results
 * if this can not be done.
 *
 * It calls clnt_create_vers_timed() with a NULL value for the timeout
 * pointer, which indicates that the default timeout should be used.
 */
CLIENT *
clnt_create_vers(const char *hostname, rpcprog_t prog, rpcvers_t *vers_out,
	rpcvers_t vers_low, rpcvers_t vers_high, const char *nettype)
{

	return (clnt_create_vers_timed(hostname, prog, vers_out, vers_low,
				vers_high, nettype, NULL));
}
#ifdef NETAPP
/* burt 395040 */
CLIENT *
clnt_create_vers_vserver(const char *hostname, rpcprog_t prog, 
			 rpcvers_t *vers_out, rpcvers_t vers_low, 
			 rpcvers_t vers_high, const char *nettype, 
			 const struct rpc_vserver_info* vsinfo)
{

	return (clnt_create_vers_timed_vserver(hostname, prog, vers_out, 
					       vers_low, vers_high, nettype, 
					       NULL, vsinfo));
}
#endif /* NETAPP */

/*
 * This the routine has the same definition as clnt_create_vers(),
 * except it takes an additional timeout parameter - a pointer to
 * a timeval structure.  A NULL value for the pointer indicates
 * that the default timeout value should be used.
 */
CLIENT *
clnt_create_vers_timed(const char *hostname, rpcprog_t prog,
    rpcvers_t *vers_out, rpcvers_t vers_low, rpcvers_t vers_high,
    const char *nettype, const struct timeval *tp)
#ifdef NETAPP
/* burt 395040 */
{
	/* will use vserver from proc struct */
	return (clnt_create_vers_timed_vserver(hostname, prog, vers_out, 
					       vers_low, vers_high, nettype, 
					       tp, NULL));
}

CLIENT *
clnt_create_vers_timed_vserver(const char *hostname, rpcprog_t prog,
			       rpcvers_t *vers_out, rpcvers_t vers_low, 
			       rpcvers_t vers_high, const char *nettype, 
			       const struct timeval *tp, 
			       const struct rpc_vserver_info* vsinfo)
#endif /* NETAPP */
{
	CLIENT *clnt;
	struct timeval to;
	enum clnt_stat rpc_stat;
	struct rpc_err rpcerr;

#ifdef NETAPP
/* burt 395040 */
	clnt = clnt_create_timed_vserver(hostname, prog, vers_high, nettype, 
					 tp, vsinfo);
#else /* !NETAPP */
	clnt = clnt_create_timed(hostname, prog, vers_high, nettype, tp);
#endif /* NETAPP */
	if (clnt == NULL) {
		return (NULL);
	}
	to.tv_sec = 10;
	to.tv_usec = 0;
	rpc_stat = clnt_call(clnt, NULLPROC, (xdrproc_t)xdr_void,
			(char *)NULL, (xdrproc_t)xdr_void, (char *)NULL, to);
	if (rpc_stat == RPC_SUCCESS) {
		*vers_out = vers_high;
		return (clnt);
	}
	while (rpc_stat == RPC_PROGVERSMISMATCH && vers_high > vers_low) {
		unsigned int minvers, maxvers;

		clnt_geterr(clnt, &rpcerr);
		minvers = rpcerr.re_vers.low;
		maxvers = rpcerr.re_vers.high;
		if (maxvers < vers_high)
			vers_high = maxvers;
		else
			vers_high--;
		if (minvers > vers_low)
			vers_low = minvers;
		if (vers_low > vers_high) {
			goto error;
		}
		CLNT_CONTROL(clnt, CLSET_VERS, (char *)&vers_high);
		rpc_stat = clnt_call(clnt, NULLPROC, (xdrproc_t)xdr_void,
				(char *)NULL, (xdrproc_t)xdr_void,
				(char *)NULL, to);
		if (rpc_stat == RPC_SUCCESS) {
			*vers_out = vers_high;
			return (clnt);
		}
	}
	clnt_geterr(clnt, &rpcerr);

error:
	rpc_createerr.cf_stat = rpc_stat;
	rpc_createerr.cf_error = rpcerr;
	clnt_destroy(clnt);
	return (NULL);
}

/*
 * Top level client creation routine.
 * Generic client creation: takes (servers name, program-number, nettype) and
 * returns client handle. Default options are set, which the user can
 * change using the rpc equivalent of _ioctl()'s.
 *
 * It tries for all the netids in that particular class of netid until
 * it succeeds.
 * XXX The error message in the case of failure will be the one
 * pertaining to the last create error.
 *
 * It calls clnt_create_timed() with the default timeout.
 */
CLIENT *
clnt_create(const char *hostname, rpcprog_t prog, rpcvers_t vers,
    const char *nettype)
{

	return (clnt_create_timed(hostname, prog, vers, nettype, NULL));
}
#ifdef NETAPP
/* burt 395040 */
CLIENT *
clnt_create_vserver(const char *hostname, rpcprog_t prog, rpcvers_t vers,
		    const char *nettype, const struct rpc_vserver_info* vsinfo)
{

	return (clnt_create_timed_vserver(hostname, prog, vers, nettype, NULL, 
					  vsinfo));
}
#endif /* NETAPP */

/*
 * This the routine has the same definition as clnt_create(),
 * except it takes an additional timeout parameter - a pointer to
 * a timeval structure.  A NULL value for the pointer indicates
 * that the default timeout value should be used.
 *
 * This function calls clnt_tp_create_timed().
 */
CLIENT *
clnt_create_timed(const char *hostname, rpcprog_t prog, rpcvers_t vers,
    const char *netclass, const struct timeval *tp)
#ifdef NETAPP
/* burt 395040 */
{
	return (clnt_create_timed_vserver(hostname, prog, vers, netclass,
					  tp, NULL));
}

CLIENT *
clnt_create_timed_vserver(const char *hostname, rpcprog_t prog, rpcvers_t vers,
			  const char *netclass, const struct timeval *tp,
			  const struct rpc_vserver_info* vsinfo)
#endif /* NETAPP */
{
	struct netconfig *nconf;
	CLIENT *clnt = NULL;
	void *handle;
	enum clnt_stat	save_cf_stat = RPC_SUCCESS;
	struct rpc_err	save_cf_error;
	char nettype_array[NETIDLEN];
	char *nettype = &nettype_array[0];

	if (netclass == NULL)
		nettype = NULL;
	else {
		size_t len = strlen(netclass);
		if (len >= sizeof (nettype_array)) {
			rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
			return (NULL);
		}
		strcpy(nettype, netclass);
	}

	if ((handle = __rpc_setconf((char *)nettype)) == NULL) {
		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
		return (NULL);
	}
	rpc_createerr.cf_stat = RPC_SUCCESS;
	while (clnt == NULL) {
		if ((nconf = __rpc_getconf(handle)) == NULL) {
			if (rpc_createerr.cf_stat == RPC_SUCCESS)
				rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
			break;
		}
#ifdef CLNT_DEBUG
		printf("trying netid %s\n", nconf->nc_netid);
#endif
#ifdef NETAPP
/* burt 395040 */
		if (vsinfo) {
#ifndef sockSetVserverContext_TASK
			unsigned int old_ctx = get_vcontext();
			(void)set_vcontext(vsinfo->vsid);
#endif
			clnt = clnt_tp_create_timed_vserver(hostname, prog, vers, nconf, tp, vsinfo);
			(void)set_vcontext(old_ctx);
		} else {
			clnt = clnt_tp_create_timed_vserver(hostname, prog, vers, nconf, tp, vsinfo);
		}
#else /* !NETAPP */
		clnt = clnt_tp_create_timed(hostname, prog, vers, nconf, tp);
#endif /* NETAPP */
		if (clnt)
			break;
		else
			/*
			 *	Since we didn't get a name-to-address
			 *	translation failure here, we remember
			 *	this particular error.  The object of
			 *	this is to enable us to return to the
			 *	caller a more-specific error than the
			 *	unhelpful ``Name to address translation
			 *	failed'' which might well occur if we
			 *	merely returned the last error (because
			 *	the local loopbacks are typically the
			 *	last ones in /etc/netconfig and the most
			 *	likely to be unable to translate a host
			 *	name).  We also check for a more
			 *	meaningful error than ``unknown host
			 *	name'' for the same reasons.
			 */
			if (rpc_createerr.cf_stat != RPC_N2AXLATEFAILURE &&
			    rpc_createerr.cf_stat != RPC_UNKNOWNHOST) {
				save_cf_stat = rpc_createerr.cf_stat;
				save_cf_error = rpc_createerr.cf_error;
			}
	}

	/*
	 *	Attempt to return an error more specific than ``Name to address
	 *	translation failed'' or ``unknown host name''
	 */
	if ((rpc_createerr.cf_stat == RPC_N2AXLATEFAILURE ||
				rpc_createerr.cf_stat == RPC_UNKNOWNHOST) &&
					(save_cf_stat != RPC_SUCCESS)) {
		rpc_createerr.cf_stat = save_cf_stat;
		rpc_createerr.cf_error = save_cf_error;
	}
	__rpc_endconf(handle);
	return (clnt);
}

/*
 * Generic client creation: takes (servers name, program-number, netconf) and
 * returns client handle. Default options are set, which the user can
 * change using the rpc equivalent of _ioctl()'s : clnt_control()
 * It finds out the server address from rpcbind and calls clnt_tli_create().
 *
 * It calls clnt_tp_create_timed() with the default timeout.
 */
CLIENT *
clnt_tp_create(const char *hostname, rpcprog_t prog, rpcvers_t vers,
    const struct netconfig *nconf)
{

	return (clnt_tp_create_timed(hostname, prog, vers, nconf, NULL));
}
#ifdef NETAPP
/* burt 395040 */
CLIENT *
clnt_tp_create_vserver(const char *hostname, rpcprog_t prog, rpcvers_t vers,
		       const struct netconfig *nconf, 
		       const struct rpc_vserver_info* vsinfo)
{

	return (clnt_tp_create_timed_vserver(hostname, prog, vers, nconf, NULL,
					     vsinfo));
}
#endif /* NETAPP */

/*
 * This has the same definition as clnt_tp_create(), except it
 * takes an additional parameter - a pointer to a timeval structure.
 * A NULL value for the timeout pointer indicates that the default
 * value for the timeout should be used.
 */
CLIENT *
clnt_tp_create_timed(const char *hostname, rpcprog_t prog, rpcvers_t vers,
    const struct netconfig *nconf, const struct timeval *tp)
#ifdef NETAPP
/* burt 395040 */
{
	/* will use vserver from proc struct */
	return (clnt_tp_create_timed_vserver(hostname, prog, vers, nconf, tp,
					     NULL));
}
CLIENT *
clnt_tp_create_timed_vserver(const char *hostname, rpcprog_t prog, 
			     rpcvers_t vers, const struct netconfig *nconf, 
			     const struct timeval *tp, 
			     const struct rpc_vserver_info* vsinfo)
#endif /* NETAPP */
{
	struct netbuf *svcaddr;			/* servers address */
	CLIENT *cl = NULL;			/* client handle */

	if (nconf == NULL) {
		rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
		return (NULL);
	}

	/*
	 * Get the address of the server
	 */
#ifdef NETAPP
	if ((svcaddr = __rpcb_findaddr_timed(prog, vers,
					     (struct netconfig *)nconf, 
					     (char *)hostname, &cl, 
					     (struct timeval *)tp, 
					     vsinfo)) == NULL) {
		/* appropriate error number is set by rpcbind libraries */
		return (NULL);
	}
	if (cl == NULL) {
		cl = clnt_tli_create_vserver(RPC_ANYFD, nconf, svcaddr,
					     prog, vers, 0, 0, vsinfo);
	} else {
#else /* !NETAPP */
	if ((svcaddr = __rpcb_findaddr_timed(prog, vers,
			(struct netconfig *)nconf, (char *)hostname,
			&cl, (struct timeval *)tp)) == NULL) {
		/* appropriate error number is set by rpcbind libraries */
		return (NULL);
	}
	if (cl == NULL) {
		cl = clnt_tli_create(RPC_ANYFD, nconf, svcaddr,
					prog, vers, 0, 0);
	} else {
		/* Reuse the CLIENT handle and change the appropriate fields */
		if (CLNT_CONTROL(cl, CLSET_SVC_ADDR, (void *)svcaddr) == TRUE) {
			if (cl->cl_netid == NULL)
				cl->cl_netid = strdup(nconf->nc_netid);
			if (cl->cl_tp == NULL)
				cl->cl_tp = strdup(nconf->nc_device);
			(void) CLNT_CONTROL(cl, CLSET_PROG, (void *)&prog);
			(void) CLNT_CONTROL(cl, CLSET_VERS, (void *)&vers);
		} else {
#endif /* NETAPP */
			CLNT_DESTROY(cl);
#ifdef NETAPP
			cl = clnt_tli_create_vserver(RPC_ANYFD, nconf, svcaddr,
						     prog, vers, 0, 0, vsinfo);
#else /* !NETAPP */
			cl = clnt_tli_create(RPC_ANYFD, nconf, svcaddr,
					prog, vers, 0, 0);
		}
#endif /* NETAPP */
	}
	free(svcaddr->buf);
	free(svcaddr);
	return (cl);
}

/*
 * Generic client creation:  returns client handle.
 * Default options are set, which the user can
 * change using the rpc equivalent of _ioctl()'s : clnt_control().
 * If fd is RPC_ANYFD, it will be opened using nconf.
 * It will be bound if not so.
 * If sizes are 0; appropriate defaults will be chosen.
 */
CLIENT *
clnt_tli_create(int fd, const struct netconfig *nconf,
	struct netbuf *svcaddr, rpcprog_t prog, rpcvers_t vers,
	uint sendsz, uint recvsz)
#ifdef NETAPP
/* burt 395040 */
{
	return (clnt_tli_create_vserver(fd, nconf, svcaddr, prog, vers,
					sendsz, recvsz, NULL));
}

CLIENT *
clnt_tli_create_vserver(int fd, const struct netconfig *nconf,
			struct netbuf *svcaddr, rpcprog_t prog, rpcvers_t vers,
			uint sendsz, uint recvsz, 
			const struct rpc_vserver_info* vsinfo)
#endif /* NETAPP */
{
	CLIENT *cl;			/* client handle */
	bool_t madefd = FALSE;		/* whether fd opened here */
	long servtype;
	int one = 1;
	struct __rpc_sockinfo si;
	extern int __rpc_minfd;

	if (fd == RPC_ANYFD) {
		if (nconf == NULL) {
			rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
			return (NULL);
		}

#ifdef NETAPP
		if (vsinfo != NULL) {
			int old_vsid = get_vcontext();
			if ((set_vcontext_socket(vsinfo->vsid)) != 0) {
				goto err;;
			}
			fd = __rpc_nconf2fd(nconf);
			restore_vcontext_socket(old_vsid);
		} else {
			fd = __rpc_nconf2fd(nconf);
		}
#else /* !NETAPP */
		fd = __rpc_nconf2fd(nconf);
#endif /* NETAPP */

		if (fd == -1)
			goto err;
		if (fd < __rpc_minfd)
			fd = __rpc_raise_fd(fd);
		madefd = TRUE;
		servtype = nconf->nc_semantics;
		if (!__rpc_fd2sockinfo(fd, &si))
			goto err;
#ifdef NETAPP
		/* Set the supplied lif role info into the socket*/
		if (__rpc_setlifrole(fd, vsinfo) != 0) {
		    goto err;
		}
                /* Ignore the error, in case application cannot set the 
                 * reserved port socket option at least they can get 
                 * a non-reserved port */
		(void)setsockopt_resvport(fd);
#else /* !NETAPP */
		bindresvport(fd, NULL);
#endif /* NETAPP */
	} else {
		if (!__rpc_fd2sockinfo(fd, &si))
			goto err;
		servtype = __rpc_socktype2seman(si.si_socktype);
		if (servtype == -1) {
			rpc_createerr.cf_stat = RPC_UNKNOWNPROTO;
			return (NULL);
		}
	}

	if (si.si_af != ((struct sockaddr *)svcaddr->buf)->sa_family) {
		rpc_createerr.cf_stat = RPC_UNKNOWNHOST;	/* XXX */
		goto err1;
	}

	switch (servtype) {
	case NC_TPI_COTS:
		cl = clnt_vc_create(fd, svcaddr, prog, vers, sendsz, recvsz);
		break;
	case NC_TPI_COTS_ORD:
		if (nconf && ((strcmp(nconf->nc_protofmly, "inet") == 0))) {
			_setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &one,
			    sizeof (one));
		}
		cl = clnt_vc_create(fd, svcaddr, prog, vers, sendsz, recvsz);
		break;
	case NC_TPI_CLTS:
		cl = clnt_dg_create(fd, svcaddr, prog, vers, sendsz, recvsz);
		break;
	default:
		goto err;
	}

	if (cl == NULL)
		goto err1; /* borrow errors from clnt_dg/vc creates */
	if (nconf) {
		cl->cl_netid = strdup(nconf->nc_netid);
		cl->cl_tp = strdup(nconf->nc_device);
	} else {
		cl->cl_netid = "";
		cl->cl_tp = "";
	}
	if (madefd) {
		(void) CLNT_CONTROL(cl, CLSET_FD_CLOSE, NULL);
/*		(void) CLNT_CONTROL(cl, CLSET_POP_TIMOD, NULL);  */
	};

	return (cl);

err:
	rpc_createerr.cf_stat = RPC_SYSTEMERROR;
	rpc_createerr.cf_error.re_errno = errno;
err1:	if (madefd)
		(void)_close(fd);
	return (NULL);
}

/*
 *  To avoid conflicts with the "magic" file descriptors (0, 1, and 2),
 *  we try to not use them.  The __rpc_raise_fd() routine will dup
 *  a descriptor to a higher value.  If we fail to do it, we continue
 *  to use the old one (and hope for the best).
 */
int __rpc_minfd = 3;

int
__rpc_raise_fd(int fd)
{
	int nfd;

	if (fd >= __rpc_minfd)
		return (fd);

	if ((nfd = _fcntl(fd, F_DUPFD, __rpc_minfd)) == -1)
		return (fd);

	if (_fsync(nfd) == -1) {
		_close(nfd);
		return (fd);
	}

	if (_close(fd) == -1) {
		/* this is okay, we will syslog an error, then use the new fd */
		(void) syslog(LOG_ERR,
			"could not close() fd %d; mem & fd leak", fd);
	}

	return (nfd);
}
#ifdef NETAPP
/*
 * This code keeps a global list of all active common_kmod RPC clients.
 * This is for debugging purposes only.
 */
struct client_list_entry {
	struct client_list_entry *prev, *next;
	CLIENT *clnt;
} *client_list_head = NULL;
int64_t client_list_len = 0;
uint64_t client_list_adds = 0;
uint64_t client_list_deletes = 0;
uint64_t client_list_enomem = 0;
uint64_t client_list_notfound = 0;

void rpc_add_client(CLIENT *clnt)
{
	struct client_list_entry *entry;

	entry = (struct client_list_entry*)malloc(sizeof(struct client_list_entry));
	if (entry == NULL) {
		/*
		 * The other counters are already protected by mutex;
		 * atomic is less risk than adding mutex here.
		 */
		atomic_add_64(&client_list_enomem, 1);
		return;
	}

	entry->prev = NULL;
	entry->clnt = clnt;

	mutex_lock(&client_list_lock);
	entry->next = client_list_head;
	if (client_list_head) {
		client_list_head->prev = entry;
	}
	client_list_head = entry;
	client_list_len++;
	client_list_adds++;
	mutex_unlock(&client_list_lock);
}

void rpc_delete_client(CLIENT *clnt)
{
	struct client_list_entry *entry;

	mutex_lock(&client_list_lock);

	entry = client_list_head;
	while (entry) {
		if (entry->clnt == clnt) {
			break;
		}
		entry = entry->next;
	}
	if (entry == NULL) {
		client_list_notfound++;
		mutex_unlock(&client_list_lock);
		return;
	}
	if (entry->prev) {
		entry->prev->next = entry->next;
	} else {
		client_list_head = entry->next;
	}
	if (entry->next) {
		entry->next->prev = entry->prev;
	}
	client_list_len--;
	assert (client_list_len >= 0);
	client_list_deletes++;
	entry->next = entry->prev = NULL;
	mutex_unlock(&client_list_lock);
	free(entry);
}
#endif /* NETAPP */
/* NetApp Footer:
 *
 * SHA512: 943f8b62e343f787258f607b15e5991d00ee3dc4340a3d3a8a20d0d8cb50bc4e7438a931d6d7b670d25c4f35c2ba6cc17fc2e2e9c203f1ecc76d234d956c1280
 * Updated: 2014/10/20 17:48:30
 */
